@@ -3,7 +3,8 @@
 The "zernpy" package is intended for initialization and calculation attributes / properties of Zernike polynomials.
 
 @author: Sergei Klykov
-@licence: MIT, @year: 2023
+@licence: MIT
+@year: 2023
 """
 if __name__ == "__main__":
     # use absolute imports for importing as module
@@ -11,7 +12,8 @@ if __name__ == "__main__":
 elif __name__ == "zernpy":
     pass  # do not add module "zernikepol" to __all__ attribute, because it demands to construct explicit path
 
-# Automatically bring the main class to the name space when the used command both commands:
+# Automatically bring the main class and some methods to the name space when the used command both commands:
 # 1) from zernpy import ZernPol; 2) from zernpy import *
-if __name__ != "__main__":
+if __name__ != "__main__" and __name__ != "__mp_main__":
     from .zernikepol import ZernPol
+    from .zernikepol import generate_polynomials
