@@ -2,6 +2,8 @@
 """
 Main script with the class definition for accessing Zernike polynomial initialization, calculation and plotting.
 
+Also, provides a few functions useful for fitting set of Zernike polynomials to an image with phases.
+
 @author: Sergei Klykov, @year: 2023 \n
 @licence: MIT \n
 
@@ -13,16 +15,19 @@ import warnings
 import math
 from collections import namedtuple
 import matplotlib.pyplot as plt
+import random
 
 # %% Local (package-scoped) imports
 if __name__ == "__main__" or __name__ == Path(__file__).stem or __name__ == "__mp_main__":
     from calculations.calc_zernike_pol import (normalization_factor, radial_polynomial, triangular_function,
                                                triangular_derivative, radial_derivative)
     from plotting.plot_zerns import plot_sum_fig, subplot_sum_on_fig
+    from calculations.fit_zernike_pols import crop_phases_img, fit_zernikes
 else:
     from .calculations.calc_zernike_pol import (normalization_factor, radial_polynomial, triangular_function,
                                                 triangular_derivative, radial_derivative)
     from .plotting.plot_zerns import plot_sum_fig, subplot_sum_on_fig
+    from .calculations.fit_zernike_pols import crop_phases_img, fit_zernikes
 
 # %% Module parameters
 __docformat__ = "numpydoc"
@@ -797,7 +802,7 @@ class ZernPol:
         return polar_vectors(R, Theta)
 
     @staticmethod
-    def plot_zernike_polynomial(polynomial):
+    def plot_zernike_polynomial(polynomial, color_map: str = "coolwarm", show_title: bool = True):
         """
         Plot the provided Zernike polynomial (instance of ZernPol class) on the matplotlib figure.
 
@@ -807,6 +812,10 @@ class ZernPol:
         ----------
         polynomial : ZernPol
             Instance of ZernPol class.
+        color_map : str, optional
+            Color map of the polar plot, recommended values: coolwarm, jet, turbo, rainbow. The default is "coolwarm".
+        show_title : bool, optional
+            Toggle for showing the name of polynomial on the plot or not.
 
         Returns
         -------
@@ -817,7 +826,11 @@ class ZernPol:
             r, theta = ZernPol.gen_polar_coordinates()
             amplitudes = [1.0]; zernikes = [polynomial]  # for reusing the sum function of polynomials
             zern_surface = ZernPol.sum_zernikes(amplitudes, zernikes, r, theta, get_surface=True)
-            plot_sum_fig(zern_surface, r, theta, title=polynomial.get_polynomial_name())
+            if show_title:
+                plot_sum_fig(zern_surface, r, theta, title=polynomial.get_polynomial_name(),
+                             color_map=color_map)
+            else:
+                plot_sum_fig(zern_surface, r, theta, "", color_map)
 
     @staticmethod
     def gen_zernikes_surface(coefficients: list, polynomials: list, r_step: float = 0.01,
@@ -853,12 +866,12 @@ class ZernPol:
     @staticmethod
     def plot_sum_zernikes_on_fig(coefficients: list, polynomials: list, figure: plt.Figure,
                                  use_defaults: bool = True, zernikes_sum_surface: zernikes_surface = (),
-                                 show_range: bool = True) -> plt.Figure:
+                                 show_range: bool = True, color_map: str = "coolwarm") -> plt.Figure:
         """
-        Plot a sum of the specified Zernike polynomials by two lists (ZernPol instances with their coefficients) on the provided figure.
+        Plot a sum of the specified Zernike polynomials by input lists (see function parameters) on the provided figure.
 
-        Note that for showing the plotted figure, one needs to call appropriate functions (e.g., matplotlib.pyplot.show() or figure.show())
-        as the method of the input parameter figure.
+        Note that for showing the plotted figure, one needs to call appropriate functions (e.g., matplotlib.pyplot.show()
+        or figure.show()) as the method of the input parameter figure.
 
         Parameters
         ----------
@@ -877,6 +890,8 @@ class ZernPol:
             Check the method signature for details. The default is ().
         show_range : bool, optional
             Flag for showing range of provided values as the colorbar on the figure. The default is True.
+        color_map : str, optional
+            Color map of the polar plot, recommended values: coolwarm, jet, turbo, rainbow. The default is "coolwarm".
 
         Raises
         ------
@@ -898,10 +913,11 @@ class ZernPol:
             zernikes_sum = ZernPol.sum_zernikes(coefficients, polynomials, polar_vectors.R, polar_vectors.Theta,
                                                 get_surface=True)
             figure = subplot_sum_on_fig(figure, zernikes_sum, polar_vectors.R, polar_vectors.Theta,
-                                        show_range_colorbar=show_range)
+                                        show_range_colorbar=show_range, color_map=color_map)
         else:
             figure = subplot_sum_on_fig(figure, zernikes_sum_surface.ZernSurf, zernikes_sum_surface.R,
-                                        zernikes_sum_surface.Theta, show_range_colorbar=show_range)
+                                        zernikes_sum_surface.Theta, show_range_colorbar=show_range,
+                                        color_map=color_map)
         return figure
 
     @staticmethod
@@ -999,16 +1015,183 @@ class ZernPol:
                 angles = float(angles)  # attempt to convert to float number, will raise ValueError if it's impossible
         # Checking that angles lie in the range [0, 2*pi]
         if isinstance(angles, np.ndarray):
-            if np.max(angles) + abs(np.min(angles)) > 2.0*np.pi:
-                warnings.warn("Theta angles defined in range outside of interval [0, 2*pi]")
+            if np.max(angles) - np.min(angles) > 2.0*np.pi:
+                _warn_message_ = "Theta angles defined in range outside of interval [0.0, 2.0*pi]"
+                warnings.warn(_warn_message_)
             elif np.max(angles) > 2.0*np.pi or np.min(angles) < 0.0:
-                warnings.warn("Max or min of theta angles lies outside of interval [0, 2*pi]")
+                _warn_message_ = "Max or min of theta angles lies outside of interval [0.0, 2.0*pi]"
+                warnings.warn(_warn_message_)
         elif isinstance(angles, float):
             if angles < 0.0 or angles > 2.0*np.pi:
-                warnings.warn("Max or min of theta angles lies outside of interval [0, 2*pi]")
+                _warn_message_ = "Max or min of theta angles lies outside of interval [0.0, 2.0*pi]"
+                warnings.warn(_warn_message_)
         return angles
 
 
+# %% Methods defs.
+def generate_polynomials(max_order: int = 10) -> tuple:
+    """
+    Generate tuple with ZernPol instances - ultimately, tuple with Zernike polynomials.
+
+    Parameters
+    ----------
+    max_order : int, optional
+        Maximum overall radial order (n) for generated Zernike list (pyramid). The default is 10.
+
+    Raises
+    ------
+    ValueError
+        Raised if max_order < 1, because it should be not less than 0.
+
+    Returns
+    -------
+    tuple
+        It composes generated ZernPol instances (Zernike polynomials).
+
+    """
+    # Sanity check of max_order parameter
+    if not isinstance(max_order, int):
+        __warning_mess = "The parameter max_order provided not as integer, there will be attempt to convert it to int"
+        warnings.warn(__warning_mess)
+        max_order = int(max_order)
+    if max_order < 0:
+        raise ValueError("The maximum order should be not less than 0")
+    if max_order > 30:
+        __warning_mess = "Calculation polynomial values with orders higher than 30 is really slow"
+        warnings.warn(__warning_mess)
+    polynomials_list = []
+    for order in range(1, max_order + 1):  # going through all specified orders
+        m = -order  # azimuthal order
+        n = order  # radial order
+        for _ in range(order + 1):  # number of polynomials = order + 1
+            polynomials_list.append(ZernPol(azimuthal_order=m, radial_order=n))
+            m += 2  # according to the specification of Zernike polynomial
+    return tuple(polynomials_list)
+
+
+def generate_random_phases(max_order: int = 4, img_width: int = 513, img_height: int = 513,
+                           round_digits: int = 4) -> tuple:
+    """
+    Generate phases image (profile) for random set of polynomials with randomly selected amplitudes.
+
+    Parameters
+    ----------
+    max_order : int, optional
+        Maximum radial order of generated Zernike polynomials. The default is 4.
+    img_width : int, optional
+        Width of generated image. The default is 513.
+    img_height : int, optional
+        Height of generated image. The default is 513.
+    round_digits : int, optional
+        Round digits for polynomials amplitudes generation (numpy.round(...) function call). The default is 4.
+
+    Returns
+    -------
+    tuple
+        Consisting of:
+        2D phase profile image with provided width and height;
+        1D numpy.ndarray containing randomly selected polynomials amplitudes;
+        tuple with generated Zernike polynomials.
+
+    """
+    # Generate list with radial and angular orders for identifying polynomials
+    polynomials_list = list(generate_polynomials(max_order))
+    # Generate list with non-zero polynomials
+    if max_order < 3:
+        selector_list = [False, False, True]
+    elif 3 <= max_order <= 4:
+        selector_list = [False, False, False, True]
+    elif 4 < max_order <= 8:
+        selector_list = [False, False, False, False, False, True]
+    else:
+        selector_list = [False, False, False, False, False, False, False, True]
+    polynomials_amplitudes = np.zeros(shape=(len(polynomials_list, )))
+    for i in range(polynomials_amplitudes.shape[0]):
+        if random.choice(selector_list):
+            polynomials_amplitudes[i] = random.uniform(-1.5, 1.5)
+    polynomials_amplitudes = np.round(polynomials_amplitudes, round_digits)
+    # Additional check that at least some amplitude is non-zero and sufficiently high
+    if np.max(np.abs(polynomials_amplitudes)) < 0.08:
+        list_indices = [i for i in range(polynomials_amplitudes.shape[0])]
+        index = random.choice(list_indices)
+        rand_ampl = random.uniform(-1.5, 1.5)
+        if abs(rand_ampl) < 0.08:  # if selected amplitude again is low, shift choice borders
+            polynomials_amplitudes[index] = random.uniform(0.25, 1.5)
+        else:
+            polynomials_amplitudes[index] = rand_ampl
+    # Generate some phases image - sum of polynomials on some 2D array of pixels converted to polar coordinates
+    phases_image = np.zeros(shape=(img_height, img_width))  # blank image
+    row_center = img_height // 2; cols_center = img_width // 2
+    # Below - define radius of the Zernike circular profile, it is +1 for including more pixels in profile
+    min_img_size = min(img_width, img_height); img_radius = 1 + min_img_size // 2
+    center = np.asarray([row_center, cols_center])  # center point of an image
+    # Calculation of 2D image with phases, which actually are composed by the sum of randomly selected Zernike polynomials
+    for i in range(img_height):
+        r = np.zeros(shape=(img_width, )); theta = np.zeros(shape=(img_width, ))
+        for j in range(img_width):
+            euclidean_dist = np.linalg.norm(center - np.asarray([i, j]))
+            if euclidean_dist <= img_radius:
+                r[j] = euclidean_dist / img_radius
+                theta[j] = np.arctan2(row_center - i, j - cols_center)
+                if theta[j] < 0.0:
+                    theta[j] += 2.0*np.pi
+        # Speed up calculations by using vectors (r, theta) as the input parameters
+        polynomials_ampl_list = polynomials_amplitudes.tolist()
+        phases_image[i, :] = ZernPol.sum_zernikes(polynomials_ampl_list,
+                                                  polynomials_list, r, theta)
+    # Final conversion
+    polynomials_list = tuple(polynomials_list)
+    return phases_image, polynomials_amplitudes, polynomials_list
+
+
+def fit_polynomials(phases_image: np.ndarray, polynomials: tuple, crop_radius: float = 1.0,
+                    suppress_warnings: bool = False, strict_circle_border: bool = False,
+                    round_digits: int = 4, return_cropped_image: bool = False) -> tuple:
+    """
+    Fit provided Zernike polynomials (instances of ZernPol class) to the 2D phase image.
+
+    Parameters
+    ----------
+    phases_image : numpy.ndarray
+        2D image with recorded phases which should be approximated by the sum of Zernike polynomials.
+    polynomials : tuple
+        Initialized tuple with instances of the ZernPol class that effectively represents target set of Zernike polynomials.
+    crop_radius : float, optional
+        Allow cropping pixel from range [0.5, 1.0], where 1.0 corresponds to radius of circle = min image size.
+        The default is 1.0.
+    suppress_warnings : bool, optional
+        Flag for suppress warnings about the provided 2D image sizes. The default is False.
+    strict_circle_border : bool, optional
+        Flag for controlling how the border pixels (on the circle radius) are treated: strictly or less strict for
+        allowing more pixels to be treated as belonged to a cropped circle. The default is False.
+    round_digits : int, optional
+        Round digits for returned polynomials amplitudes (numpy.round(...) function call). The default is 4.
+    return_cropped_image : bool, optional
+        Flag for calculating and returning cropped image, used for fitting procedure. The default is False.
+
+    Returns
+    -------
+    tuple
+        Depending on the input parameter (flag) "return_cropped_image":
+        if it is True, then tuple returned with following variables: zernike_coefficients - 1D numpy.ndarray
+        with the fitted coefficients of Zernike polynomials, and cropped_image - the cropped part from the
+        input image with phases that is used for fitting procedure (useful for debugging purposes);
+        if it is False, the following tuple will be returned: zernike_coefficients, None - 1st with the same
+        meaning and type as explained before.
+    """
+    zernike_coefficients = np.zeros(shape=(len(polynomials), ))
+    logic_mask, polar_coordinates = crop_phases_img(phases_image, crop_radius, suppress_warnings,
+                                                    strict_circle_border)
+    if return_cropped_image:
+        cropped_image = logic_mask*phases_image  # for debugging
+    zernike_coefficients = fit_zernikes(phases_image, logic_mask, polar_coordinates, polynomials)
+    zernike_coefficients = np.round(zernike_coefficients, round_digits)
+    if return_cropped_image:
+        return zernike_coefficients, cropped_image
+    else:
+        return zernike_coefficients, None
+
+
 # %% Test functions for the external call
 def check_conformity():
     """
@@ -1085,7 +1268,7 @@ def check_conformity():
 if __name__ == "__main__":
     check_conformity()  # testing initialization
     plt.close("all")  # Check plotting functions
-    # zp = ZernPol(m=0, n=4); ZernPol.plot_zernike_polynomial(zp)  # basic plotting
+    zp = ZernPol(m=0, n=4); ZernPol.plot_zernike_polynomial(zp, color_map="jet", show_title=False)  # basic plotting
     # Two plots should look similar below, using different methods to call
     zp1 = ZernPol(m=0, n=2); zp2 = ZernPol(m=-3, n=3); coefficients = [1.0, 1.0]; polynomials = [zp1, zp2]
     # fig = plt.figure(figsize=(4, 4)); fig = ZernPol.plot_sum_zernikes_on_fig(coefficients, polynomials, fig)
@@ -1093,9 +1276,14 @@ if __name__ == "__main__":
     # fig1 = ZernPol.plot_sum_zernikes_on_fig(coefficients, polynomials, fig1, show_range=False)
     # ZernPol._plot_zernikes_half_pyramid()
     fig3 = plt.figure(figsize=(2, 2)); zp3 = ZernPol(osa=9); polynomials = [zp3]; coefficients = [1.0]
-    fig3 = ZernPol.plot_sum_zernikes_on_fig(coefficients, polynomials, fig3, show_range=False)
-    fig3.subplots_adjust(0, 0, 1, 1); plt.show()
+    fig3 = ZernPol.plot_sum_zernikes_on_fig(coefficients, polynomials, fig3, show_range=False, color_map="turbo")
+    fig3.subplots_adjust(0, 0, 1, 1)
     # Check that warning is raised
-    z = ZernPol(n=26, m=-10); print(z.radial(0.85))
-    # Check that warning not raised
-    ZernPol(n=32, m=32); ZernPol(n=28, m=-26)
+    # z = ZernPol(n=26, m=-10); print("Value of radial Zernike pol. with n=26, m=-10 r=0.85:", round(z.radial(0.85), 4))
+    # Tests with generation / restoring Zernike profiles (phases images)
+    phases_image, polynomials_ampls, polynomials = generate_random_phases(img_height=301, img_width=301)
+    plt.figure(); plt.axis("off"); plt.imshow(phases_image, cmap="jet"); plt.tight_layout(); plt.subplots_adjust(0, 0, 1, 1)
+    polynomials_amplitudes, cropped_img = fit_polynomials(phases_image, polynomials, return_cropped_image=True,
+                                                          strict_circle_border=False, crop_radius=1.0)
+    plt.figure(); plt.axis("off"); plt.imshow(cropped_img, cmap="jet"); plt.tight_layout(); plt.subplots_adjust(0, 0, 1, 1)
+    plt.show()  # show all images created by plt.figure() calls
